package othello.observation;

import java.util.List;
import java.util.ArrayList;
import java.rmi.* ;
import java.rmi.server.UnicastRemoteObject ;

/** A class that informs observers potentially asynchronously.
 *  Each observer is sent an object of class CountMonitorInterface
 *  and is expected to call "decrement" on that method exactly once.
 *  Once all the observers have called decrement, the notifying object
 *  (the object of this class) proceeds.
 * 
 * <p>Title: Othello</p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Memorial University</p>
 * @author Theodore S. Norvell
 * @version 1.0
 */

public class RemotelyConcurrentlyObservable extends UnicastRemoteObject {
	protected RemotelyConcurrentlyObservable() throws RemoteException {
		super();
	}

	private ReadersAndWriters rw = new ReadersAndWriters() ;
	
    private static final long serialVersionUID = 4205794150994351885L;
    
	private List<RemoteConcurrentObserver> observerList
        = new ArrayList<RemoteConcurrentObserver>() ;
    
    protected void readerEnters () { rw.readerEnters() ; }
    
    protected void readerExits() { rw.readerExits() ; }
    
    protected void writerEnters() { rw.writerEnters() ; }
    
    protected void writerExits() { rw.writerExits() ; }
    
    protected void writerGoesPassive() { rw.writerGoesPassive() ; }
    
    protected void writerGoesActive() { rw.writerGoesActive() ; }
    
    public void addObserver( RemoteConcurrentObserver observer ) {
    	writerEnters() ;
    	try {
    		System.err.println("Adding observer "+observer);
    		observerList.add( observer ); }
    	finally{ writerExits() ; } }
        
    /** Notify all observers of a change.
     * This routine should only be called by a writer.
     * Observers may read, but must not write.
     * @param arg
     */
    protected void notifyObservers( Object arg ) {
    	writerGoesPassive() ;
        System.err.println("Notifying observers");
        RemoteCountMonitor countMonitor ; 
        try {
            countMonitor = new RemoteCountMonitor( observerList.size() ) ; }
        catch( RemoteException e ) {
            // This shouldn't happen. Panic
            countMonitor = null ;
            System.err.println("Unexpected failure to create RemoteCountMonitor object" ) ;
            e.printStackTrace( System.err );
            System.exit(1) ; }
        for( RemoteConcurrentObserver observer : observerList) {
            try { 
                System.err.println("Observer notified...");
                observer.update( countMonitor, arg ) ;
                System.err.println("...ok");}
            catch( RemoteException e ) {
                // If call to update fails, we'll assume that
                // the remote observer did not get the message and decrement
                // the counter
                System.err.println("...unsuccessfully") ;
                e.printStackTrace(System.err);
                countMonitor.decrement() ; } }
        countMonitor.waitForZero() ;
        writerGoesActive() ; }
}

class RemoteCountMonitor
extends UnicastRemoteObject
implements RemoteCountMonitorInterface {
    
	private static final long serialVersionUID = -7852891784027321638L;
	
	private int count ;
    
    RemoteCountMonitor( int start ) throws RemoteException {
        super() ;
        count = start ; }
    
    public synchronized void decrement() {
        count -= 1 ;
        if( count == 0 ) notifyAll() ; }
    
    synchronized void waitForZero() {
        while( count > 0 ) 
            try { wait() ; }
            catch( InterruptedException e ) {
            	Thread.currentThread().interrupt() ; } }
}
